
/*
 *      writed by kir http://paradox.org.ua
 */

#include <sys/types.h>
#include <sys/errno.h>
#include <sys/param.h>
#include <sys/conf.h>
#include <sys/kernel.h>
#include <sys/bus.h>

#include <sys/selinfo.h>

#include "dvbapi.h"

#include "devio.h"
#include "ttbudget.h"

static int dev_num = 0;

static d_open_t         dvb_open;
static d_close_t        dvb_close;
static d_read_t         dvb_read;
static d_write_t        dvb_write;
static d_ioctl_t        dvb_ioctl;
static d_mmap_t         dvb_mmap;
static d_poll_t         dvb_poll;
static d_kqfilter_t	dvb_kqfilter;

static struct cdevsw dvb_cdevsw = {
	.d_version =    D_VERSION,
	.d_flags =      D_NEEDGIANT,
	.d_open =       dvb_open,
	.d_close =      dvb_close,
	.d_read =       dvb_read,
	.d_write =      dvb_write,
	.d_ioctl =      dvb_ioctl,
	.d_poll =       dvb_poll,
	.d_mmap =       dvb_mmap,
	.d_kqfilter =	dvb_kqfilter,
	.d_name =       "dvb"
};

static void	filt_dvbdetach(struct knote *kn);
static int	filt_dvbread(struct knote *kn, int hint);
static int	filt_dvbwrite(struct knote *kn, int hint);

static struct filterops	dvb_read_filterops = {
	.f_isfd =	1,
	.f_attach =	NULL,
	.f_detach =	filt_dvbdetach,
	.f_event =	filt_dvbread,
};

static struct filterops	dvb_write_filterops = {
	.f_isfd =	1,
	.f_attach =	NULL,
	.f_detach =	filt_dvbdetach,
	.f_event =	filt_dvbwrite,
};

void
dvb_init(struct devcallback *dvbdev)
{
	knlist_init(&dvbdev->kq_sel.si_note, NULL, NULL, NULL, NULL);

	dvbdev->dev = make_dev(&dvb_cdevsw,
				dev_num,
				UID_ROOT,
				GID_WHEEL,
				0600,
				"dvb%d", dev_num);

	dev_num ++;
}

void
dvb_free(struct devcallback *dvbdev)
{
	destroy_dev(dvbdev->dev);

	knlist_destroy(&dvbdev->kq_sel.si_note);
}


#define UNIT(x)         ((x) & 0x0f)
#define FUNCTION(x)     (x >> 4)

static int
dvb_open(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	struct adapter *hw;
	int unit;
	int err;

	unit = UNIT(minor(dev));

	hw = devclass_get_softc(ttdevclass, unit);
	if (hw == NULL)
	{
		return ENXIO;
	}

	device_busy(devclass_get_device(ttdevclass, unit));

	switch (FUNCTION(minor(dev)) )
	{

	}

	return 0;
}

static int
dvb_close(struct cdev *dev, int flags, int fmt, struct thread *td)
{
	struct adapter *hw;
	int unit;
	int err;

	unit = UNIT(minor(dev));

	hw = devclass_get_softc(ttdevclass, unit);
	if (hw == NULL)
	{
		return ENXIO;
	}

	device_unbusy(devclass_get_device(ttdevclass, unit));

	switch (FUNCTION(minor(dev)) )
	{

	}

	return 0;
}

static int
dvb_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	return ENXIO;
}

static int
dvb_write(struct cdev *dev, struct uio *uio, int ioflag)
{
	return ENXIO;
}

static int
dvb_ioctl(struct cdev *dev, unsigned long cmd, caddr_t arg, int flag, struct thread *td)
{
	struct adapter *hw;
	int unit;
	int err;

	unit = UNIT(minor(dev));

	hw = devclass_get_softc(ttdevclass, unit);
	if (hw == NULL)
	{
		return ENXIO;
	}

	err = 0;

	switch (cmd)
	{
	case DVBAPI:
		{
			struct dvbapi *parm = (struct dvbapi *)arg;

			if (hw->dvbdev.devControl != NULL && hw->dvbdev.data != NULL)
			{
				if (hw->dvbdev.devControl(hw->dvbdev.data, parm) != 0)
				{
					err = EINVAL;
				} else
				{
					err = 0;
				}
			} else
				err = EINVAL;
		}
			break;

	default:
		printf("%s: unsupported ioctl %08x!\n", __FUNCTION__, (unsigned long)cmd);
		err = EINVAL;
	}

	return err;
}

static int
dvb_mmap(struct cdev *dev, vm_offset_t offset, vm_paddr_t *paddr, int nprot)
{
	return ENXIO;
}

static int
dvb_poll(struct cdev *dev, int events, struct thread *tp)
{
	return ENXIO;
}

static int
dvb_kqfilter(struct cdev *dev, struct knote *kn)
{
	int unit;
	struct adapter *hw;

	unit = UNIT(minor(dev));

	hw = devclass_get_softc(ttdevclass, unit);

	switch (kn->kn_filter)
	{
	case EVFILT_READ:
		kn->kn_fop = &dvb_read_filterops;
		break;
	case EVFILT_WRITE:
		kn->kn_fop = &dvb_write_filterops;
		break;

	default:
		return EINVAL;
	}

	kn->kn_hook = hw;

	knlist_add(&hw->dvbdev.kq_sel.si_note, kn, 0);

	return 0;
}

static void
filt_dvbdetach(struct knote *kn)
{
	struct adapter *hw = kn->kn_hook;

	knlist_remove(&hw->dvbdev.kq_sel.si_note, kn, 0);
}

static int
filt_dvbread(struct knote *kn, int hint)
{
	struct adapter *hw = kn->kn_hook;

	return 0;
}

static int
filt_dvbwrite(struct knote *kn, int hint)
{
	struct adapter *hw = kn->kn_hook;

	return 0;
}
