
/*
 *      writed by kir http://paradox.org.ua
 */

extern "C"
{
	#include <sys/param.h>

	#include <sys/time.h>

	#include "timeout.h"
}

#include "pci.h"
#include "i2c.h"

CTTSpciI2C::CTTSpciI2C(CTTSpciIoIrq *pci):pci(pci)
{
	Init();
}

CTTSpciI2C::~CTTSpciI2C()
{
}

void
CTTSpciI2C::Sleep(unsigned int ms)
{
	/*
	TODO
	*/
}

void
CTTSpciI2C::Init(void)
{
	ErrNo = 0;
	Error = 0;

	IICTRF = 0;
	Count = 0;

	Buffer[0] = (int *)NULL;
	Buffer[1] = (int *)NULL;
	Buffer[2] = (int *)NULL;

	IICSTATE = 0x600;	//speed 320bs

	pci->SetReg(0xFC, 0x01000100, 1);

	pci->SetReg(0x8C, 0, 0);
	pci->SetReg(0x90, IICSTATE, 1);

	Slave = 0;
}

int
CTTSpciI2C::GetSlave(void)
{
	return Slave;
}

void
CTTSpciI2C::SetSlave(int slave)
{
	Slave = slave;
}

void
CTTSpciI2C::SetIICTRF(int data, unsigned char cmd, unsigned char index)
{
	unsigned char dataIndex;
	unsigned char cmdIndex;

	dataIndex = 24 - (index * 8);
	cmdIndex = 6 - (index * 2);

	IICTRF &= ~((3 << cmdIndex) | ( 0xff << dataIndex));

	IICTRF |= (data & 0xff) << dataIndex;
	IICTRF |= (cmd & 0xff) << cmdIndex;
}

void
CTTSpciI2C::ClearIICTRF(void)
{
	IICTRF = 0;
	Count = 0;
}

void
CTTSpciI2C::ReadIICTRF(void)
{
	if (Buffer[0] != NULL)
	{
		*Buffer[0] = (IICTRF >> 24 ) & 0xff;
		Buffer[0] = (int *)NULL;
	}

	if (Buffer[1] != NULL)
	{
		*Buffer[1] = (IICTRF >> 16 ) & 0xff;
		Buffer[1] = (int *)NULL;
	}

	if (Buffer[2] != NULL)
	{
		*Buffer[2] = (IICTRF >> 8 ) & 0xff;
		Buffer[2] = (int *)NULL;
	}
}

int
CTTSpciI2C::CheckIICSTATE(void)
{
	int error;

	error = pci->GetReg(0x90);

	if ( !(IICTRF & 2) && !(error & 0x7E))
	{
		return 1;
	}

	Error |= (error & 0x7C);

	pci->SetReg(0x90, IICSTATE | 0x80, 1);
	pci->SetReg(0x90, IICSTATE, 1);

	error = pci->GetReg(0x90);

	if (error & 0x7F)
		pci->SetReg(0x90, IICSTATE, 1);

	ErrNo++;

	return 0;
}

void
CTTSpciI2C::FlushIICTRF(void)
{
	int status;
	struct timeval timeout;

	Sleep(0);

	Error = 0;

	if (pci->GetReg(0x90) & 3)
	{
		pci->SetReg(0x8C, 0, 0);
		pci->SetReg(0x90, IICSTATE | 0x80, 1);
		pci->SetReg(0x90, IICSTATE, 1);

		Sleep(0);

		if (pci->GetReg(0x90) & 3)
			pci->SetReg(0x90, IICSTATE, 1);
	}

	/* write 3bytes */
	pci->SetReg(0x8C, IICTRF, 1);

	SetTimeOut(&timeout, 100);	//100
	Sleep(0);

	status = pci->GetReg(0x100);
	if ( !(status & 1) )
	{
		while (GetTimeOut(&timeout))
		{
			Sleep(0);
			status = pci->GetReg(0x100);
			if ( status & 1 )
				break;
		}

		if ( !(status & 1) )
		{
			Error |= 1;
			IICTRF = 0xFFFFFF02;
			return;
		}
	}

	/* read 3bytes */
	SetTimeOut(&timeout, 100);	//100
	Sleep(0);

	IICTRF = pci->GetReg(0x8C);
	if ( IICTRF & 1 )
	{
		while (GetTimeOut(&timeout))
		{
			Sleep(0);
			IICTRF = pci->GetReg(0x8C);
			if ( !(IICTRF & 1) )
				break;
		}

		if ( IICTRF & 3 )
		{
			Error |= 1;
			IICTRF = 0xFFFFFF02;
			return;
		}
	}

	Sleep(0);
}

void
CTTSpciI2C::AddIICTRF(int *data, unsigned char cmd, int res)
{
	Buffer[Count] = (res == 1) ? data : 0;// NULL;

	SetIICTRF(*data, cmd, Count++);

	if (Count == 3)
	{
		FlushIICTRF();
		ReadIICTRF();
		CheckIICSTATE();
		ClearIICTRF();
	}
}

int
CTTSpciI2C::ReadSeq(int *readBuf, int readLen)
{
	int slave;
	int errNo;
	int i;

	ClearIICTRF();

	slave = (Slave << 1) | 1;
	errNo = (ErrNo);

	AddIICTRF(&slave, 3, 0);	//START WRITE

	for (i = 0; i < readLen-1; i++)
	{
		AddIICTRF(&readBuf[i], 2, 1);	//COUNT READ
		if (errNo < ErrNo)
			return Error;
	}

	AddIICTRF(&readBuf[i], 1, 1);	//STOP READ

	if (Count != 0)
		FlushIICTRF();

	if (CheckIICSTATE())
		ReadIICTRF();

	return Error;
}

int
CTTSpciI2C::WriteSeq(int *writeBuf, int writeLen)
{
	int slave;
	int errNo;
	int i;

	ClearIICTRF();

	slave = (Slave << 1);
	errNo = (ErrNo);

	AddIICTRF(&slave, 3, 0);	//START WRITE

	for (i = 0; i < writeLen-1; i++)
	{
		AddIICTRF(&writeBuf[i], 2, 0);	//COUNT WRITE
		if (errNo < ErrNo)
			return Error;
	}

	AddIICTRF(&writeBuf[i], 1, 0);	//STOP WRITE

	if (Count != 0)
		FlushIICTRF();

	CheckIICSTATE();

	return Error;
}

int
CTTSpciI2C::CombinedSeq(int *writeBuf, int writeLen, int *readBuf, int readLen)
{
	int slave;
	int errNo;
	int i;

	ClearIICTRF();

	slave = (Slave << 1);	//
	errNo = (ErrNo);

	AddIICTRF(&slave, 3, 0);	//START WRITE

	for (i = 0; i < writeLen; i++)
	{
		AddIICTRF(&writeBuf[i], 2, 0);	//COUNT WRITE
		if (errNo < ErrNo)
			return Error;
	}

	slave = (Slave << 1) | 1;

	AddIICTRF(&slave, 3, 0);	//I2C_START WRITE
	if (errNo < ErrNo)
		return Error;

	for (i = 0; i < readLen-1; i++)
	{
		AddIICTRF(&readBuf[i], 2, 1);	//I2C_COUNT READ
		if (errNo < ErrNo)
			return Error;
	}

	AddIICTRF(&readBuf[i], 1, 1);	//I2C_STOP READ

	if (Count != 0)
		FlushIICTRF();

	if(CheckIICSTATE())
		ReadIICTRF();

	return Error;
}

